# 2^31 or UINT_MAX/2 rounded up
TRANSITION_TARGET_TABLE_START = 2147483648
# this is hopefully the same as static_cast<float>(UINT_MAX) in C++
INFINITE_WEIGHT = float(4294967295)
# USHRT_MAX
NO_SYMBOL_NUMBER = 65535
NO_TABLE_INDEX = 4294967295
